#!/usr/bin/env python3
# Importing libraries
import argparse
import fcntl
import os
import sys

from pynmrstar import Entry

CHAIN_CODE = "chain_code"

SEQUENCE_CODE = "sequence_code"

parser = None

EXIT_ERROR = 1


def exit_error(msg):
    if parser:
        parser.print_help()
    print()
    print(f"ERROR: {msg}")
    print("exiting...")
    sys.exit(EXIT_ERROR)


def offset_residue_numbers(entry, chain, offset):
    for frame_data in entry.frame_dict.values():
        for loop_data in frame_data.loop_dict.values():
            if SEQUENCE_CODE in loop_data.tags:
                sequence_index = loop_data.tag_index(SEQUENCE_CODE)
                chain_index = loop_data.tag_index(CHAIN_CODE)
                for line in loop_data:
                    if line[chain_index] == chain:
                        sequence = line[sequence_index]
                        if "@" not in sequence:
                            try:
                                sequence = int(sequence)
                            except ValueError:
                                pass
                        if isinstance(sequence, int):
                            sequence = sequence + offset
                            line[sequence_index] = str(sequence)


def read_args():
    global parser
    parser = argparse.ArgumentParser(
        description="Add chemical shift errors to a NEF file"
    )
    parser.add_argument(
        "-c",
        "--chain",
        metavar="CHAIN",
        type=str,
        default="A",
        dest="chain",
        help="chain in which to offset residue numbers",
    )
    parser.add_argument(
        "-o",
        "--offset",
        metavar="OFFSET",
        type=int,
        default=0,
        dest="offset",
        help="offset to add to residue numbers",
    )
    parser.add_argument(metavar="FILES", nargs=argparse.REMAINDER, dest="files")

    return parser.parse_args()


def check_stream():
    # make stdin a non-blocking file
    fd = sys.stdin.fileno()
    fl = fcntl.fcntl(fd, fcntl.F_GETFL)
    fcntl.fcntl(fd, fcntl.F_SETFL, fl | os.O_NONBLOCK)

    try:
        result = sys.stdin.read()
    except Exception:
        result = ""

    return result


if __name__ == "__main__":

    args = read_args()

    is_tty = sys.stdin.isatty()

    if is_tty and len(args.files) == 0:
        parser.print_help()
        print()
        msg = """I require at least 1 argument or input stream with a chemical_shift_list frame"""

        exit_error(msg)

    entries = []
    try:
        if len(args.files) == 0:
            lines = check_stream()
            if len(lines) != 0:
                entries.append(Entry.from_string(lines))
            else:
                exit_error("Error: input appears to be empty")
        else:
            for file in args.files:
                entries.append(Entry.from_file(file))
    except OSError as e:
        msg = f"couldn't open target nef file because {e}"
        exit_error(msg)

    for entry in entries:
        offset_residue_numbers(entry, args.chain, args.offset)

        # print(entry)



import io
import os
import sys
import traceback
from argparse import Namespace
from pathlib import Path
from typing import Dict, Iterator, List, Optional, TextIO, TypeVar, Union

from cacheable_iter import iter_cache
from pynmrstar import Entry, Loop, Saveframe

from nef_pipelines.lib.constants import (
    EXIT_ERROR,
    NEF_META_DATA,
    NEF_PIPELINES,
    NEF_PIPELINES_VERSION,
    NEF_UNKNOWN,
)
from nef_pipelines.lib.header_lib import (
    create_header_frame,
    get_creation_time,
    get_uuid,
)
from nef_pipelines.lib.structures import LineInfo


def _get_loop_by_category_or_none(frame: Saveframe, category: str) -> Loop:

    result = None
    if f"_{category}" in frame.loop_dict.keys():
        result = frame.get_loop(category)

    return result


def loop_add_data_tag_dict(loop: Loop, data: Dict[str, object]) -> None:
    tagged_data = []
    for category_tag in loop.get_tag_names():
        _, tag = category_tag.split(".")
        if tag in data:
            tagged_data.append(data[tag])
        else:
            tagged_data.append(NEF_UNKNOWN)

    loop.add_data(tagged_data)


# TODO: there is no space for scrript arguments!
def fixup_metadata(entry: Entry, name: str, version: str, script: str):
    """
    add a new entry to a nef metadata frame

    Args:
        entry (Entry): the target entry
        name (str): the name of the program used
        version (str): the vserion of the program used
        script (str): the script used

    Returns:
        Entry: the modified entry
    """

    if entry is not None and NEF_META_DATA in entry.category_list:
        meta_frame = entry[NEF_META_DATA]

        last_program = meta_frame.get_tag("program_name")[0]
        last_program_version = meta_frame.get_tag("program_version")[0]
        last_script_name = meta_frame.get_tag("script_name")[0]

        meta_frame.add_tag("program_name", name, update=True)
        meta_frame.add_tag("program_version", version, update=True)
        meta_frame.add_tag("script_name", script, update=True)
        creation_time = get_creation_time()
        meta_frame.add_tag("creation_time", creation_time, update=True)
        uuid = get_uuid(NEF_PIPELINES, creation_time)
        meta_frame.add_tag("uuid", uuid, update=True)

        run_history_loop = _get_loop_by_category_or_none(meta_frame, "nef_run_history")
        if run_history_loop is not None:
            if run_history_loop.get_tag("run_number"):
                run_number_tags = run_history_loop.get_tag("run_number")
                run_numbers = [int(run_number) for run_number in run_number_tags]
                last_run_number = max(run_numbers)
                next_run_number = last_run_number + 1
            else:
                next_run_number = 1

            data = {
                "run_number": next_run_number,
                "program_name": last_program,
                "program_version": last_program_version,
                "script_name": last_script_name,
            }
            loop_add_data_tag_dict(run_history_loop, data)
        else:
            run_history_loop = Loop.from_scratch("nef_run_history")
            run_history_loop.add_tag(
                ["run_number", "program_name", "program_version", "script_name"]
            )
            run_history_loop.add_data(["."] * 4)
    else:
        header = create_header_frame(NEF_PIPELINES, NEF_PIPELINES_VERSION, script)
        entry.add_saveframe(header)


class StringIteratorIO(io.TextIOBase):

    # https://stackoverflow.com/questions/12593576/adapt-an-iterator-to-behave-like-a-file-like-object-in-python
    # second answer with more votes!

    def __init__(self, iter):
        self._iter = iter
        self._left = ""

    def readable(self):
        return True

    def _read1(self, n=None):
        while not self._left:
            try:
                self._left = next(self._iter)
            except StopIteration:
                break
        ret = self._left[:n]
        self._left = self._left[len(ret) :]
        return ret

    def read(self, n=None):
        line = []
        if n is None or n < 0:
            while True:
                m = self._read1()
                if not m:
                    break
                line.append(m)
        else:
            while n > 0:
                m = self._read1(n)
                if not m:
                    break
                n -= len(m)
                line.append(m)
        return "".join(line)

    def readline(self):
        line = []
        while True:
            i = self._left.find("\n")
            if i == -1:
                line.append(self._left)
                try:
                    self._left = next(self._iter)
                except StopIteration:
                    self._left = ""
                    break
            else:
                line.append(self._left[: i + 1])
                self._left = self._left[i + 1 :]
                break
        return "".join(line)


# https://stackoverflow.com/questions/6657820/how-to-convert-an-iterable-to-a-stream/20260030#20260030
def iterable_to_stream(iterable, buffer_size=io.DEFAULT_BUFFER_SIZE):
    """
    Lets you use an iterable (e.g. a generator) that yields bytestrings as a read-only
    input stream.

    The stream implements Python 3's newer I/O API (available in Python 2's io module).
    For efficiency, the stream is buffered.
    """

    return StringIteratorIO(iterable)


def running_in_pycharm():
    return "PYCHARM_HOSTED" in os.environ


def get_pipe_file(args: Namespace) -> Optional[TextIO]:
    """
    get an input on stdin or from an argument called pipe

    Args:
        args (Namespace): command line arguments

    Returns:
        TextIO: an input stream or None

    """

    result = None
    if "pipe" in args and args.pipe:
        result = iter(cached_file_stream(args.pipe))
    # pycharm doesn't treat stdstreams correcly and hangs
    elif not sys.stdin.isatty() and not running_in_pycharm():
        result = cached_stdin()

    return StringIteratorIO(result) if result else None


def get_pipe_file_or_exit(args: Namespace) -> Optional[TextIO]:

    try:
        result = get_pipe_file(args)
    except Exception as e:
        exit_error("couldn't read from stdin or -pipe file", e)

    if isinstance(result, StringIteratorIO):

        result = "\n".join(list(result))

        if len(result) == 0:
            result = None
        else:
            result = io.StringIO(result)

    if result is None:
        exit_error("couldn't read from stdin and no -pipe in command line arguments")

    return result


@iter_cache
def cached_stdin():
    # TODO: this doesn't work!'
    return sys.stdin


class cached_file_stream:
    def __init__(self, file_name):
        self._file_name = file_name

    def __enter__(self):
        return _cached_file_stream(self._file_name)

    def __exit__(self, *args):
        pass

    def __iter__(self):
        return _cached_file_stream(self._file_name)


@iter_cache
def _cached_file_stream(file_name):
    try:
        with open(file_name, "r") as lines:
            result = lines.readlines()
    except IOError as e:
        exit_error(f"couldn't open stream {file_name} because {e}")
    return result


def script_name(file: str) -> Path:
    """
    get the name of the script

    Args:
        file (str): the name of the file

    Returns:
        Path: path to the file
    """
    return Path(file).name


def exit_error(msg, exception=None):
    """
    print an error message and exit error

    Args:
        msg: the message
    """

    if exception is not None:
        exc_info = sys.exc_info()
        traceback.print_exception(*exc_info, file=sys.stderr)
        print(file=sys.stderr)

    print(f"ERROR: {msg}", file=sys.stderr)
    print("exiting...", file=sys.stderr)
    sys.exit(EXIT_ERROR)


def process_stream_and_add_frames(
    frames: List[Saveframe], input_args: Namespace
) -> Entry:
    """
    take a set of save frames and either add them to a stream from stdin / a pipe-file or create
    a new stream with a proper NEF metadata header

    Args:
        frames: a set of save frames
        input_args: command line arguments for an entry_name and pipe file source

    Returns:
        a new entry containing the frames
    """

    try:
        stream = get_pipe_file(input_args)
    except Exception as e:
        exit_error(f"failed to load pipe file because {e}")

    if stream is not None:
        lines = "".join(stream.readlines())
    else:
        lines = ""

    if len(lines.strip()) == 0:
        stream = None

    new_entry = (
        Entry.from_string(lines)
        if stream
        else Entry.from_scratch(input_args.entry_name)
    )

    fixup_metadata(
        new_entry, NEF_PIPELINES, NEF_PIPELINES_VERSION, script_name(__file__)
    )

    # TODO: check if frame exists
    for frame in frames:
        new_entry.add_saveframe(frame)

    return new_entry


def is_int(value: str) -> bool:
    """
    Check if a string is an integer
    Args:
        value (str): the putative integer

    Returns bool:
        true is if its an integer

    """
    result = False
    try:
        int(value)
        result = True
    except ValueError:
        pass

    return result


def is_float(value: str) -> bool:
    """
    Check if a string is an float
    Args:
        value (str): the putative float

    Returns bool:
        true is if its an integer

    """
    result = False
    try:
        float(value)
        result = True
    except ValueError:
        pass

    return result


T = TypeVar("T")


# https://stackoverflow.com/questions/312443/how-do-i-split-a-list-into-equally-sized-chunks
def chunks(input: Iterator[T], n: int) -> Iterator[List[T]]:
    """Yield successive n-sized chunks from lst.
       lst: the list to chunk
       n: the chunk size

    Returns:
        an iterator of chunks from lst of length T

    """
    lst = list(input)
    for i in range(0, len(lst), n):
        yield lst[i : i + n]


def _build_int_float_error_message(message, line_info, field, format):
    messages = []
    if message:
        messages.append(message)
    if line_info:
        message = f"""
                    couldn't convert {line_info.line} to format

                    at {line_info.line_no + 1} in f{line_info.file_name}

                    line value was

                    {line_info.line}

                    field was

                    {field}
                """
        messages.append(message)
    message = "\n\n".join(messages)
    return message


def read_float_or_exit(
    string: str, line_info: LineInfo = None, field="unknown", message=None
) -> float:
    """
    convert a string to an float or exit with error including line information

    Args:
        string: string to parse
        line_info: the line the string came from
        field: the name of the field defininbg the string (ignored if there is no line info)
        message: any other messages to show

    Returns:
        a float
    """

    format = "float"
    try:
        result = float(string)
    except Exception:
        message = _build_int_float_error_message(message, line_info, field, format)

        exit_error(message)

    return result


def read_integer_or_exit(
    string: str, line_info: LineInfo = None, field="unknown", message=None
) -> int:
    """
    convert a string to an int or exit with error including line information

    Args:
        string: string to parse
        line_info: the line the string came from
        field: the name of the field defining the string (ignored if there is no line info)
        message: any other messages to show

    Returns:
        an integer
    """

    format = "integer"
    try:
        result = int(string)
    except Exception:

        message = _build_int_float_error_message(message, line_info, field, format)

        exit_error(message)

    return result


# TODO test this!
def parse_comma_separated_options(
    lists: Union[List[List[str]], List[str], str]
) -> List[str]:
    """
    Take a mixed list of strings or strings that can be parsed as a comma separated list of strings
    and make a list of all the items

    e.g.

    test = [['1','2','3'], 'a,b,c', ['4','5']]

    gives

    ['1','2','3','a','b','c','4','5']

    :param lists: a mixture of  lists of string or comma separated as a at sting
    :return: list of items
    """
    result = []
    for item in lists:
        if isinstance(item, str):
            item = item.strip(",")
            result.extend(item.split(","))
        else:
            result.append(item)

    return result



from random import seed

import pytest


@pytest.fixture
def typer_app():
    import typer

    import nef_app

    if not nef_app.app:
        nef_app.app = typer.Typer()

        # if only one command is registered it is ignored on the command line...
        dummy_app = typer.Typer()
        nef_app.app.add_typer(dummy_app, name="dummy")

    return nef_app.app


@pytest.fixture
def fixed_seed():
    seed(42)


@pytest.fixture
def clear_cache():
    """
    clear the lru cache used by lib.util cached_stdin
    see https://stackoverflow.com/questions/40273767/clear-all-lru-cache-in-python
    """
    import functools
    import gc

    gc.collect()
    wrappers = [
        a for a in gc.get_objects() if isinstance(a, functools._lru_cache_wrapper)
    ]

    for wrapper in wrappers:
        wrapper.cache_clear()



from lib.structures import  ShiftList, ShiftData, AtomLabel
from textwrap import dedent
from transcoders.nmrview.nmrview_lib import  parse_shifts
import pytest
from lib.test_lib import assert_lines_match, isolate_frame, path_in_test_data

from typer.testing import CliRunner
runner = CliRunner()


METADATA_NMRVIEW ='nef_nmr_meta_data'
LIST_CHAINS = ['chains', 'list']


@pytest.fixture
def using_chains():
    # register the module under test
    import tools.chains

# noinspection PyUnusedLocal
def test_frame_basic(typer_app, using_chains, monkeypatch):

    monkeypatch.setattr('sys.stdin.isatty', lambda: False)

    path = path_in_test_data(__file__, 'multi_chain.nef')

    result = runner.invoke(typer_app, [*LIST_CHAINS, '--pipe', path])

    if result.exit_code != 0:
        print('INFO: stdout from failed read:\n', result.stdout)

    assert result.exit_code == 0

    EXPECTED =  'A B C'

    assert_lines_match (EXPECTED, result.stdout)


#!/usr/bin/env python3
import argparse
import sys
import operator
from array import array

from numpy import average
from pynmrstar import Entry
from dataclasses import dataclass

parser = None

EXIT_ERROR = 1
def exit_error(msg):

        print(f'ERROR: {msg}')
        print('exiting...')
        sys.exit(EXIT_ERROR)

@dataclass(frozen=True)
class Assignment:
    assignment: int
    fixed: bool = False
    merit: float = 0.0

def read_args():
    global parser
    parser = argparse.ArgumentParser(description='Assign a NEF file based on output from an assignment in NEF format')
    parser.add_argument('-t', '--target', metavar='TARGET_NEF', type=str, default=None, dest='target',
                        help='target nef file to assign')
    parser.add_argument(metavar='ASSIGNMENT', nargs=1, dest='assignment')

    return parser.parse_args()

if __name__ == '__main__':

    args = read_args()

    try:
        nef_target_data = Entry.from_file(args.target)
    except OSError as e:
        msg = f"couldn't open target nef file because {e}"
        exit_error(msg)

    try:
        res_assign = Entry.from_file(args.assignment[0])
    except OSError as e:
        msg = f"couldn't open residue assignments file because {e}"
        exit_error(msg)

    shift_list_frames = []
    for frame_name in nef_target_data.frame_dict:
        if 'nef_chemical_shift_list' in frame_name:
            shift_list_frames.append(frame_name)

    if not 'ccpn_residue_assignments' in res_assign.frame_dict:
        exit_error("res_assign doesn't contain a ccpn_residue_assignments frame")

    
    residue_assignments = {}
    residue_types = {}

    # TODO need constants for names 
    # TODO need a check of required columns
    for loop in res_assign['ccpn_residue_assignments'].loop_dict.values():

        loop_tags = {tag:i for i,tag in enumerate(loop.tags)}

        for line_number, line in enumerate(loop):
            
            fixed = line[loop_tags['fixed']].lower() in ('true', '.')

            
            # need functions to check
            chain = line[loop_tags['chain_code']]
            residue_number = int(line[loop_tags['residue_number']])
            residue_type = line[loop_tags['residue_type']]
           

            residue_key = chain,residue_number
            residue_types[residue_key] = residue_type

            if not fixed:
                assignment = int(line[loop_tags['assignment']])
                assigned_residue = int(line[loop_tags['assignment']])
                merit =  float(line[loop_tags['merit']])

                residue_assignments.setdefault(residue_key,[]).append(Assignment(assignment,False, merit))

                if residue_key in residue_types:
                    if (bad_residue_type := residue_types[residue_key]) != residue_type:
                        msg  = f"""more than one residue type for assignment
                                   line {line_number} 
                                   file {sys.argv[2]} 

                                   residue types {bad_residue_type} {residue_type}
                                """



    min_merit = 0.6
    filtered_residue_assignments = {}
    for residue_key, assignments in residue_assignments.items():
        assignments.sort(key=operator.attrgetter('merit'))

        assignment = assignments[0]
        if assignment.merit > min_merit:
            filtered_residue_assignments[residue_key] =  assignment


assignment_to_residue = {}
for residue, assignment in filtered_residue_assignments.items():
    assignment_to_residue[assignment] = residue 


nmr_residue_to_residue = {}
for assignment, residue in assignment_to_residue.items():

    nmr_residue = '@%i' % assignment.assignment
    nmr_residue_minus_1 = '@%i-1' % assignment.assignment

    nmr_residue_to_residue[nmr_residue] = (*residue, residue_types[residue])

    residue_minus_1 = (residue[0], residue[1]-1)
    nmr_residue_to_residue[nmr_residue_minus_1] = (*residue_minus_1, residue_types[residue_minus_1])

# for key, value in nmr_residue_to_residue.items():
#     print(key,value)

class TagIndexKey:
    def __init__(self, tag_indices):
        self._tags_indices = tag_indices

    def __call__(self, data):
        result = []


        for tag_index in self._tags_indices:
            value = data[tag_index]

            try:
                value = int(value)
            except ValueError:
                pass

            if isinstance(value, int):
                result.append((0, value))
            elif isinstance(value, str):
                result.append((1, value))

        return tuple(result)

target_frames = ['nef_chemical_shift_list', 'nef_nmr_spectrum']
for frame_name, frame_data in nef_target_data.frame_dict.items():


    target_type = None
    for target_frame in target_frames:
        if frame_name.startswith(target_frame):
            target_type = target_frame
            break
    # print(frame_name, target_type)

    if target_type == 'nef_chemical_shift_list':


        for loop_name, loop_data in frame_data.loop_dict.items():


            sequence_code_index = loop_data.tag_index('sequence_code')
            chain_code_index = loop_data.tag_index('chain_code')
            residue_name_index = loop_data.tag_index('residue_name')
            atom_name_index = loop_data.tag_index('atom_name')

            value_index = loop_data.tag_index('value')
            value_uncertainty_index = loop_data.tag_index('value_uncertainty')


            tagKey = TagIndexKey([chain_code_index,sequence_code_index,atom_name_index])
            for line_number, line in enumerate(loop_data):
                nmr_residue = line[sequence_code_index]

                if nmr_residue in nmr_residue_to_residue:

                    chain_code, residue_number, residue_type = nmr_residue_to_residue[nmr_residue]

                    line[chain_code_index] = chain
                    line[sequence_code_index] = residue_number
                    line[residue_name_index] = residue_type

            line_by_assignment = {}
            for line_number, line in enumerate(loop_data):
                key = line[chain_code_index], line[sequence_code_index], line[atom_name_index]
                line_by_assignment.setdefault(key,[]).append(line)

            for key, lines in line_by_assignment.items():
                if len(lines) > 1:
                    value = average([float(line[value_index]) for line in lines])
                    uncertainty = average([float(line[value_uncertainty_index]) for line in lines])


                    lines[0][value_index] =  value
                    lines[0][value_uncertainty_index] = uncertainty
                    for line in lines[1:]:
                        del loop_data.data[loop_data.data.index(line)]

                # elif '@' in nmr_residue:
                #     print(f'WARNING: shift list, nmr residue {nmr_residue} not found in file skipping')

            loop_data.sort_rows(tags=['chain_code','sequence_code','atom_name'], key = tagKey)# print(line)




    # print (nmr_residue_to_residue)
    if target_type == 'nef_nmr_spectrum':

        loop_types = {}

        for loop_name, loop_data in frame_data.loop_dict.items():
            
            if loop_name == '_nef_peak':

                loop_tags = {tag:i for i,tag in enumerate(loop_data.tags)}

                # print(loop_tags)
                
                active_indices = []
                for i in range(1,15):
                    test_name  = 'atom_name_%i' % i
                    if test_name in loop_tags:
                        active_indices.append(i)


                for data in loop_data:
                    for active_index in active_indices:
                        # print(data)
                        chain_index = loop_tags['chain_code_%i' % active_index]
                        atom_index = loop_tags['atom_name_%i' % active_index]
                        residue_number_index = loop_tags['sequence_code_%i' % active_index]
                        residue_type_index = loop_tags['residue_name_%i' % active_index]
                        
                        residue_number = data[residue_number_index]
                        if residue_number in nmr_residue_to_residue:
                            chain_code, residue_number, residue_type = nmr_residue_to_residue[residue_number]

                            data[chain_index] = chain_code
                            data[residue_number_index] = residue_number
                            data[residue_type_index] = residue_type

                        elif '@' in nmr_residue:
                            print(f'WARNING: spectrum, nmr residue {nmr_residue} not found in file skipping')    
                             
                    # print(data)
    # print(nef_target_data)
                # for active_index in active_indices:



            # tag_indices = set()
            # for tag in loop_tags:
            #     if tag.startswith('position'):
            #         tag_index = tag.split('_')[-1]
            #     # print(tag_index)
            #         tag_indices.add(tag_index)
            # # print('indices',tag_indices)
            # print(loop_data.category)

            # print(loop.tag_index)


            # sequence_code_index = loop_tags['sequence_code']
            # chain_code_index = loop_tags['chain_code']
            # residue_name_index = loop_tags['residue_name']
            # for line_number, line in enumerate(loop_data):
            #     if (nmr_residue := line[sequence_code_index]) in nmr_residue_to_residue:
            #         chain_code, residue_number, residue_type = nmr_residue_to_residue[nmr_residue]

            #         line[chain_code_index] =  chain
            #         line[sequence_code_index] = residue_number
            #         line[residue_name_index] =  residue_type

                    # line[sequence]

# should sort rows but need a schema
# for loop in nef_target_data['nef_chemical_shift_list_default']:
#     loop.sort_rows(tags =['chain_code', 'sequence_code'])
print(nef_target_data)

    # for frame_name, frame_data in nef_target_data.frame_dict.items():
    #
    #
    #
    #
    #     for loop_name, loop_data in frame_data.loop_dict.items():
    #         print(id(loop_data))


        













from lib.structures import ShiftList, ShiftData, AtomLabel, SequenceResidue
from textwrap import dedent
from transcoders.nmrpipe.nmrpipe_lib import read_db_file_records , read_shift_file
import pytest
from lib.test_lib import assert_lines_match, isolate_frame, path_in_test_data


from typer.testing import CliRunner
runner = CliRunner()

SHIFTS_NMRPIPE = 'nef_chemical_shift_list_nmrpipe'
METADATA_NMRPIPE ='nef_nmr_meta_data'
NMRPIPE_IMPORT_SHIFTS = ['nmrpipe', 'import', 'shifts']

HEADER = open(path_in_test_data(__file__,'test_header_entry.txt', local=False)).read()


@pytest.fixture
def using_nmrpipe():
    # register the module under test
    import transcoders.nmrpipe


def test_lib_parse_shifts():

    EXPECTED = ShiftList(
        shifts=[
            ShiftData(atom=AtomLabel(SequenceResidue(chain_code='A', sequence_code=1, residue_name='ALA'), atom_name='N'), shift=125.74, error=None),
            ShiftData(atom=AtomLabel(SequenceResidue(chain_code='A', sequence_code=2, residue_name='GLY'), atom_name='HN'), shift=8.31, error=None)
        ]
    )


    test_data = '''
        VARS   RESID RESNAME ATOMNAME SHIFT
        FORMAT %4d %1s %4s %8.2f

        1   ALA   N   125.74
        2   GLY   HN   8.31

    '''

    test_data = dedent(test_data)

    test_records = read_db_file_records(test_data.split('\n'))

    shifts = read_shift_file(test_records)
    assert shifts == EXPECTED

EXPECTED_NS3_S135A_BMRB1_SHORT = '''\
    save_nef_chemical_shift_list_nmrpipe
       _nef_chemical_shift_list.sf_category   nef_chemical_shift_list
       _nef_chemical_shift_list.sf_framecode  nef_chemical_shift_list_nmrpipe
    
       loop_
          _nef_chemical_shift_list.chain_code
          _nef_chemical_shift_list.sequence_code
          _nef_chemical_shift_list.residue_name
          _nef_chemical_shift_list.atom_name
          _nef_chemical_shift_list.value
          _nef_chemical_shift_list.value_uncertainty
          _nef_chemical_shift_list.element
          _nef_chemical_shift_list.isotope_number
    
         A   1   ALA   N     125.74   .   .   .    
         A   1   ALA   C     179.08   .   .   .    
         A   2   GLY   CA    44.77    .   .   .    
         A   2   GLY   HA2   3.89     .   .   .    
         A   2   GLY   HA3   3.8      .   .   .    
    
       stop_
    
    save_
'''

# noinspection PyUnusedLocal
def test_ns3_S135A_BMRB1_short(typer_app, using_nmrpipe, monkeypatch):

    monkeypatch.setattr('sys.stdin.isatty', lambda: False)

    path = path_in_test_data(__file__, 'ns3_S135A_BMRB1_short.txt')
    result = runner.invoke(typer_app, [*NMRPIPE_IMPORT_SHIFTS, path], input=HEADER)

    if result.exit_code != 0:
        print(result.stdout)

    assert result.exit_code == 0
    mol_sys_result = isolate_frame(result.stdout, '%s' % SHIFTS_NMRPIPE)

    assert_lines_match(EXPECTED_NS3_S135A_BMRB1_SHORT, mol_sys_result)


import pytest

from lib.test_lib import select_matching_tests

TEST_NAMES = [
    'tests/nmrpipe/test_gdb.py::test_bad_data_format',
    'tests/nmrpipe/test_gdb.py::test_bad_format',
    'tests/nmrview/test_tcl.py::test_basic_word',
    'tests/nmrview/test_tcl.py::test_quoted_word'
]


def test_test_match_specific():

    EXPECTED_NAMES = [
        'tests/nmrpipe/test_gdb.py::test_bad_data_format',
        'tests/nmrview/test_tcl.py::test_basic_word'
    ]

    result = select_matching_tests(TEST_NAMES, EXPECTED_NAMES)

    assert result == EXPECTED_NAMES


def test_test_match_stared_path():

    EXPECTED_NAMES = [
        'tests/nmrview/test_tcl.py::test_basic_word',
    ]

    args = ['*nmrview*::test_basic_word']

    result = select_matching_tests(TEST_NAMES, args)

    assert result == EXPECTED_NAMES


def test_test_match_stared_name():
    args = ['tests/nmrview/test_tcl.py::*']

    EXPECTED_NAMES = [
        'tests/nmrview/test_tcl.py::test_basic_word',
        'tests/nmrview/test_tcl.py::test_quoted_word'
    ]

    result = select_matching_tests(TEST_NAMES, args)

    assert result == EXPECTED_NAMES


def test_just_name():
    args = ['test_basic_word']

    EXPECTED_NAMES = [
        'tests/nmrview/test_tcl.py::test_basic_word',
    ]

    result = select_matching_tests(TEST_NAMES, args)

    assert result == EXPECTED_NAMES


def test_star():
    args = ['*']

    result = select_matching_tests(TEST_NAMES, args)

    assert result == TEST_NAMES


def test_test_match_truncated_path():

    EXPECTED_NAMES = [
        'tests/nmrpipe/test_gdb.py::test_bad_data_format',
        'tests/nmrpipe/test_gdb.py::test_bad_format',
    ]

    args = ['nmrpipe/test_gdb.py::*']
    result = select_matching_tests(TEST_NAMES, args)

    assert result == EXPECTED_NAMES


def test_no_path():
    args = ['::test_bad_format']

    EXPECTED_NAMES = [
        'tests/nmrpipe/test_gdb.py::test_bad_format',
    ]

    result = select_matching_tests(TEST_NAMES, args)

    assert result == EXPECTED_NAMES

def test_no_name():
    args = ['tests/nmrpipe/test_gdb.py::']

    EXPECTED_NAMES = [
        'tests/nmrpipe/test_gdb.py::test_bad_data_format',
        'tests/nmrpipe/test_gdb.py::test_bad_format',
    ]

    result = select_matching_tests(TEST_NAMES, args)

    assert result == EXPECTED_NAMES


def test_bare_separator():

    args = ['::']

    result = select_matching_tests(TEST_NAMES, args)

    assert result == TEST_NAMES


def test_bare_filename():

    args = ['test_gdb.py::']

    EXPECTED_NAMES = [
        'tests/nmrpipe/test_gdb.py::test_bad_data_format',
        'tests/nmrpipe/test_gdb.py::test_bad_format',
    ]

    result = select_matching_tests(TEST_NAMES, args)

    assert result == EXPECTED_NAMES


if __name__ == '__main__':
    pytest.main([__file__, '-vv'])


import sys
from argparse import Namespace
from enum import auto
from fnmatch import fnmatch
from pathlib import Path
from typing import Dict, Iterator, List, Union

# from pandas import DataFrame
from pynmrstar import Entry, Loop, Saveframe
from strenum import LowercaseStrEnum

from nef_pipelines.lib.util import (
    cached_stdin,
    exit_error,
    is_float,
    is_int,
    running_in_pycharm,
)

NEF_CATEGORY_ATTR = "__NEF_CATEGORY__"

UNUSED = "."


# what the selection type is for Star SaveFrames
class SelectionType(LowercaseStrEnum):
    NAME = auto()
    CATEGORY = auto()
    ANY = auto()

    # see https://github.com/irgeek/StrEnum/issues/9
    def _cmp_values(self, other):
        return self.value, str(other).upper()


# currently disabled as they add a dependency on pandas and numpy
# def loop_to_dataframe(loop: Loop) -> DataFrame:
#     """
#     convert a pynmrstar Loop to a pandas DataFrame. Note the Loop category is
#     saved in the dataframe's attrs['__NEF_CATEGORY__']
#
#     :param loop: the pynmrstar Loop
#     :return: a pandas DataFrame
#     """
#     tags = loop.tags
#     data = DataFrame()
#
#     # note this strips the preceding _
#     data.attrs[NEF_CATEGORY_ATTR] = loop.category[1:]
#
#     for tag in tags:
#         if tag != "index":
#             data[tag] = loop.get_tag(tag)
#
#     return data
#
#
# def dataframe_to_loop(frame: DataFrame, category: str = None) -> Loop:
#     """
#     convert a pandas DataFrame to a pynmrstar Loop
#
#     :param frame: the pandas DataFrame
#     :param category: the star category note this will override any category stored in attrs
#     :return: the new pynmrstar Loop
#     """
#     loop = Loop.from_scratch(category=category)
#     loop_data = {}
#     for column in frame.columns:
#         loop.add_tag(column)
#         loop_data[column] = list(frame[column])
#
#     loop.add_data(loop_data)
#
#     if NEF_CATEGORY_ATTR in frame.attrs and not category:
#         loop.set_category(frame.attrs[NEF_CATEGORY_ATTR])
#     elif category:
#         loop.set_category(category)
#
#     return loop


def select_frames_by_name(
    frames: Union[List[Saveframe], Entry],
    name_selectors: Union[List[str], str],
    exact=False,
) -> List[Saveframe]:
    """
    select frames  by names and wild cards, to avoid typing *s on the command line the match is greedy by default
    if an exact match is not found for one of the frames first time we search will all the name selectors turned into
    wild cards by surrounding them with the * as a fence so name_selector-> *name_selector*

    :param frames: the list of frames or entry to search
    :param name_selectors: a single string or list of strings to use to match frame names, selectors can contain
                           wild cards used by pythons fnmatch
    :param exact: do exact matching and don't search again with wildcards added if no frames are selected
    :return: a list or matching frames
    """

    def match_frames(frames, name_selectors):

        result = {}

        for frame in frames:
            for selector in name_selectors:

                if fnmatch(frame.name, selector):
                    # frames aren't hashable and so can't be saved in a set
                    # but names should be unique
                    result[frame.name] = frame

        return result

    if isinstance(name_selectors, str):
        name_selectors = [
            name_selectors,
        ]

    result = match_frames(frames, name_selectors)

    if not exact and len(result) == 0:

        name_selectors = [f"*{selector}*" for selector in name_selectors]
        result = match_frames(frames, name_selectors)

    return list(result.values())


# refactor to two functions one of which gets a TextIO
def create_entry_from_stdin_or_exit() -> Entry:

    """
    read a star file entry from stdin or exit withan error message
    :return: a star file entry
    """

    try:
        if not sys.stdin.isatty() or running_in_pycharm():
            stdin = cached_stdin()
            if cached_stdin is None:
                lines = ""
            else:
                lines = "".join(stdin)

            if len(lines.strip()) == 0:
                raise Exception("stdin is empty")
            else:
                entry = Entry.from_string(lines)
        else:
            exit_error("you appear to be reading from an empty stdin")
    except Exception as e:
        exit_error(
            f"failed to read nef entry from stdin because the NEF parser replied: {e}",
            e,
        )

    return entry


# TODO we should examine columns for types not individual rows entries
def loop_row_dict_iter(
    loop: Loop, convert: bool = True
) -> Iterator[Dict[str, Union[str, int, float]]]:
    """
    create an iterator that loops over the rows in a star file Loop as dictionaries, by default sensible
    conversions from strings to ints and floats are made
    :param loop: the Loop
    :param convert: try to convert values to ints or floats if possible [default is True]
    :return: iterator of rows as dictionaries
    """
    for row in loop:
        row = {tag: value for tag, value in zip(loop.tags, row)}

        if convert:
            for key in row:
                row[key] = do_reasonable_type_conversions(row[key])

        yield row


def do_reasonable_type_conversions(value: str) -> Union[str, float, int]:
    """
    do reasonable type conversion from str to int or float
    :param value: the string to convert
    :return: value converted from str to int or float if possible
    """
    if is_int(value):
        value = int(value)
    elif is_float(value):
        value = float(value)
    return value


def loop_row_namespace_iter(loop: Loop, convert: bool = True) -> Iterator[Namespace]:
    """
    create an iterator that loops over the rows in a star file Loop as Namespaces, by default sensible
    conversions from strings to ints and floats are made
    :param loop: thr Loop
    :param convert: try to convert values to ints or floats if possible [default is True]
    :return: iterator of rows as dictionaries
    """
    for row in loop_row_dict_iter(loop, convert=convert):
        yield Namespace(**row)


# TODO this partially overlaps with select_frames_by_name in this file, combine and simplify!
def select_frames(
    entry: Entry, selector_type: SelectionType, filters: List[str]
) -> List[Saveframe]:
    """
    select a list of frames by name of either category or name

    :param entry: the entry in which frames are looked for
    :param selector_type: the matching type frame.name or frame.category or both [default, search order
           frame.name frame.category]
    :param filters: a list of strings to use as filters as defined by fnmatch
    :return: a list of selected saveframes
    """

    star_filters = [f"*{filter}*" for filter in filters]
    filters = list(filters)
    filters.extend(star_filters)

    result = []
    for frame in entry.frame_dict.values():

        accept_frame_category = any(
            [fnmatch(frame.category, filter) for filter in filters]
        )
        accept_frame_name = any([fnmatch(frame.name, filter) for filter in filters])

        if (
            selector_type in (SelectionType.NAME, SelectionType.ANY)
            and not accept_frame_name
        ):
            continue

        if (
            selector_type in (SelectionType.CATEGORY, SelectionType.ANY)
            and not accept_frame_category
        ):
            continue

        result.append(frame)

    return result


def read_entry_from_file_or_stdin_or_exit_error(file: Path) -> Entry:
    """
    read a star entry from stdin or a file or exit.
    note 1. the stdin stream is cached
         2. this exits with an error if stdin can't be read because its a terminal

    :param file:
    :return:
    """
    if file is None:
        entry = create_entry_from_stdin_or_exit()
    else:
        try:
            with open(file) as fh:
                entry = Entry.from_file(fh)

        except IOError as e:
            exit_error(f"couldn't read from the file {file}", e)
    return entry



import pytest
from typer.testing import CliRunner

from lib.test_lib import assert_lines_match, path_in_test_data

runner = CliRunner()

DELETE_FRAMES = ["frames", "delete"]


@pytest.fixture
def using_frames():
    # register the module under test
    import tools.frames  # noqa: F401


EXPECTED_DELETE_CATEGORY = """\
        data_pales_test

        save_nef_rdc_restraint_list_test_1
           _nef_rdc_restraint_list.sf_category     nef_rdc_restraint_list
           _nef_rdc_restraint_list.sf_framecode    nef_rdc_restraint_list_test_1
           _nef_rdc_restraint_list.potential_type  log-normal

           loop_
              _nef_rdc_restraint.index
              _nef_rdc_restraint.restraint_id
              _nef_rdc_restraint.restraint_combination_id
              _nef_rdc_restraint.chain_code_1
              _nef_rdc_restraint.sequence_code_1
              _nef_rdc_restraint.residue_name_1
              _nef_rdc_restraint.atom_name_1
              _nef_rdc_restraint.chain_code_2
              _nef_rdc_restraint.sequence_code_2
              _nef_rdc_restraint.residue_name_2
              _nef_rdc_restraint.atom_name_2
              _nef_rdc_restraint.weight
              _nef_rdc_restraint.target_value
              _nef_rdc_restraint.target_value_uncertainty

             1   1   .   A   2   TRP   H   A   21   TRP   N   2   -5.2   0.33
             2   2   .   A   3   GLY   H   A   22   GLY   N   3   3.1    0.4

           stop_

        save_
    """


# noinspection PyUnusedLocal
def test_delete_type(typer_app, using_frames, monkeypatch, clear_cache):

    monkeypatch.setattr("sys.stdin.isatty", lambda: False)

    path = path_in_test_data(__file__, "pales_test_1.nef", local=True)

    input_stream = open(path).read()

    result = runner.invoke(typer_app, [*DELETE_FRAMES, "-c", "mol"], input=input_stream)

    if result.exit_code != 0:
        print("INFO: stdout from failed read:\n", result.stdout)

    assert result.exit_code == 0

    assert_lines_match(EXPECTED_DELETE_CATEGORY, result.stdout)


EXPECTED_DELETE_NAME = """\
    data_pales_test

    save_nef_molecular_system
       _nef_molecular_system.sf_category   nef_molecular_system
       _nef_molecular_system.sf_framecode  nef_molecular_system

       loop_
          _nef_sequence.index
          _nef_sequence.chain_code
          _nef_sequence.sequence_code
          _nef_sequence.residue_name
          _nef_sequence.linking
          _nef_sequence.residue_variant
          _nef_sequence.cis_peptide

         1   A   1   ALA   start    .   .
         2   A   2   TRP   middle   .   .
         3   A   3   GLY   middle   .   .

       stop_

    save_

"""


# noinspection PyUnusedLocal
def test_delete_name(typer_app, using_frames, monkeypatch, clear_cache):

    monkeypatch.setattr("sys.stdin.isatty", lambda: False)

    path = path_in_test_data(__file__, "pales_test_1.nef", local=True)

    input_stream = open(path).read()

    result = runner.invoke(typer_app, [*DELETE_FRAMES, "test_1"], input=input_stream)

    if result.exit_code != 0:
        print("INFO: stdout from failed read:\n", result.stdout)

    assert result.exit_code == 0

    assert_lines_match(EXPECTED_DELETE_NAME, result.stdout)



import pytest

from transcoders.nmrview import nmrview_lib



@pytest.fixture
def parser():

    return nmrview_lib.get_tcl_parser()

@pytest.fixture
def parse_tcl():

    return nmrview_lib.parse_tcl

def test_basic_word(parser):
    values = [
        "123",
        "abc",
        "123.abc",
        "123.abc#"
    ]

    for value in values:
        parsed = parser.parseString(value)

        assert len(parsed) == 1
        assert parsed[0] == value


def test_quoted_word(parser):
    values = [
        '"123"',
        '"abc"',
        '"123.abc"'
    ]

    for value in values:
        parsed = parser.parseString(value)

        assert len(parsed) == 1
        assert parsed[0] == value.strip('"')


def test_quoted_word_list(parser):
    values = [
        '"123 456"',
        '"abc def"',
        '"123.abc 456.def"'
    ]

    for value in values:
        parsed = parser.parseString(value)

        assert len(parsed) == 2
        assert list(parsed) == value.strip('"').split()


def test_string_and_quoted_word_list(parser):
    tests_and_expecteds = {
        '"abc" "123 456"': ['abc', ['123', '456']],
        '"" "123 456"':    ['', ['123', '456']],
    }

    for test, expected in tests_and_expecteds.items():
        parsed = parser.parseString(test)


        assert list(parsed.asList()) == expected


def test_complex_list(parser):
    tests_and_expecteds = {
        '{123 456}':          ['123', '456'],
        '"" {123 456}':       ['', ['123', '456']],
        '{} {123 456}':       ['', ['123', '456']],
        '{} {{} 123 456}':    ['', [[], '123', '456']],
        '{} {{} 123 {} 456}': ['', [[], '123', [], '456']],
        '{} "{} 123 {} 456"': ['', [[], '123', [], '456']],
    }

    for test, expected in tests_and_expecteds.items():
        parsed = parser.parseString(test)

        assert list(parsed.asList()) == expected


def test_rest_of_line(parse_tcl, capsys):
    test = '12{3'

    with pytest.raises(SystemExit) as excinfo:
        parse_tcl(test, file_name='Wibble.txt', line_no=666)

    stderr = capsys.readouterr().err
    print(stderr)
    assert "Expected end of text, found '{'" in stderr
    assert "line: 667" in stderr
    assert "file: Wibble.txt" in stderr





if __name__ == '__main__':
    pytest.main([__file__, '-vv'])



import pytest
from typer.testing import CliRunner

from lib.test_lib import assert_lines_match, isolate_frame, path_in_test_data

HEADER = open(path_in_test_data(__file__, "test_header_entry.txt", local=False)).read()

MOLECULAR_SYSTEM = "nef_molecular_system"
PDB_IMPORT_SEQUENCE = ["pdb", "import", "sequence"]

runner = CliRunner()


@pytest.fixture
def using_pdb():
    # register the module under test
    import transcoders.pdb  # noqa: F401


EXPECTED_3AA = """\
save_nef_molecular_system
   _nef_molecular_system.sf_category   nef_molecular_system
   _nef_molecular_system.sf_framecode  nef_molecular_system

   loop_
      _nef_sequence.index
      _nef_sequence.chain_code
      _nef_sequence.sequence_code
      _nef_sequence.residue_name
      _nef_sequence.linking
      _nef_sequence.residue_variant
      _nef_sequence.cis_peptide

     1   A   1   ALA   start    .   .
     2   A   2   ALA   middle   .   .
     3   A   3   ALA   end      .   .

   stop_

save_"""


# noinspection PyUnusedLocal
def test_3aa(typer_app, using_pdb, monkeypatch):

    monkeypatch.setattr("sys.stdin.isatty", lambda: False)

    path = path_in_test_data(__file__, "3aa.pdb", local=True)
    result = runner.invoke(typer_app, [*PDB_IMPORT_SEQUENCE, path], input=HEADER)

    print(result.stdout)

    assert result.exit_code == 0
    mol_sys_result = isolate_frame(result.stdout, "%s" % MOLECULAR_SYSTEM)

    assert_lines_match(EXPECTED_3AA, mol_sys_result)


EXPECTED_3A_AB = """\
save_nef_molecular_system
   _nef_molecular_system.sf_category   nef_molecular_system
   _nef_molecular_system.sf_framecode  nef_molecular_system

   loop_
      _nef_sequence.index
      _nef_sequence.chain_code
      _nef_sequence.sequence_code
      _nef_sequence.residue_name
      _nef_sequence.linking
      _nef_sequence.residue_variant
      _nef_sequence.cis_peptide

     1   A    1   ALA   start    .   .
     2   A    2   ALA   middle   .   .
     3   A    3   ALA   end      .   .
     4   B   11   ALA   start    .   .
     5   B   12   ALA   middle   .   .
     6   B   13   ALA   end      .   .

   stop_

save_"""


# noinspection PyUnusedLocal
def test_3a_ab(typer_app, using_pdb, monkeypatch):

    monkeypatch.setattr("sys.stdin.isatty", lambda: False)

    path = path_in_test_data(__file__, "3a_ab.pdb")
    result = runner.invoke(typer_app, [*PDB_IMPORT_SEQUENCE, path], input=HEADER)

    assert result.exit_code == 0

    mol_sys_result = isolate_frame(result.stdout, "%s" % MOLECULAR_SYSTEM)

    assert_lines_match(EXPECTED_3A_AB, mol_sys_result)


EXPECTED_3A_CCCC_DDDD = """\
save_nef_molecular_system
   _nef_molecular_system.sf_category   nef_molecular_system
   _nef_molecular_system.sf_framecode  nef_molecular_system

   loop_
      _nef_sequence.index
      _nef_sequence.chain_code
      _nef_sequence.sequence_code
      _nef_sequence.residue_name
      _nef_sequence.linking
      _nef_sequence.residue_variant
      _nef_sequence.cis_peptide

     1   CCCC    1   ALA   start    .   .
     2   CCCC    2   ALA   middle   .   .
     3   CCCC    3   ALA   end      .   .
     4   DDDD   11   ALA   start    .   .
     5   DDDD   12   ALA   middle   .   .
     6   DDDD   13   ALA   end      .   .

   stop_

save_"""


# noinspection PyUnusedLocal
def test_3a_segid_cccc_dddd(typer_app, using_pdb, monkeypatch):

    monkeypatch.setattr("sys.stdin.isatty", lambda: False)

    path = path_in_test_data(__file__, "3a_ab.pdb")
    result = runner.invoke(
        typer_app, [*PDB_IMPORT_SEQUENCE, "--segid", path], input=HEADER
    )

    assert result.exit_code == 0

    mol_sys_result = isolate_frame(result.stdout, "%s" % MOLECULAR_SYSTEM)

    assert_lines_match(EXPECTED_3A_CCCC_DDDD, mol_sys_result)


# noinspection PyUnusedLocal
def test_3a_force_segid_cccc_dddd(typer_app, using_pdb, monkeypatch):

    monkeypatch.setattr("sys.stdin.isatty", lambda: False)

    path = path_in_test_data(__file__, "3a_cccc_dddd.pdb")
    result = runner.invoke(typer_app, [*PDB_IMPORT_SEQUENCE, path], input=HEADER)

    print(result.stdout)

    assert result.exit_code == 0

    mol_sys_result = isolate_frame(result.stdout, "%s" % MOLECULAR_SYSTEM)

    assert_lines_match(EXPECTED_3A_CCCC_DDDD, mol_sys_result)


# noinspection PyUnusedLocal
def test_3a_test_no_chain_segid(typer_app, using_pdb, monkeypatch):

    monkeypatch.setattr("sys.stdin.isatty", lambda: False)

    path = path_in_test_data(__file__, "3a_no_chain_no_segid.pdb")
    result = runner.invoke(typer_app, [*PDB_IMPORT_SEQUENCE, path], input=HEADER)

    assert result.exit_code == 1

    assert "ERROR: residue with no chain code" in result.stdout



from dataclasses import replace
from typing import Dict, List, Tuple

import tabulate
import typer
from pynmrstar import Saveframe

from nef_pipelines.lib.nef_lib import (
    create_entry_from_stdin_or_exit,
    loop_row_dict_iter,
    select_frames_by_name,
)
from nef_pipelines.lib.sequence_lib import (
    get_sequence_or_exit,
    sequence_residues_to_sequence_3let,
    translate_3_to_1,
)
from nef_pipelines.lib.structures import AtomLabel, RdcRestraint, SequenceResidue
from nef_pipelines.lib.util import exit_error, read_float_or_exit
from nef_pipelines.transcoders.nmrpipe.nmrpipe_lib import print_pipe_sequence
from nef_pipelines.transcoders.pales import export_app

app = typer.Typer()

# TODO: name translations
# TODO: correct weights
# TODO: move utuilities to lib
# TODO: support multiple chains


# noinspection PyUnusedLocal
@export_app.command()
def rdcs(
    chains: List[str] = typer.Option(
        [],
        "-c",
        "--chain",
        help="chains to export, to add mutiple chains use repeated calls  [default: 'A']",
        metavar="<CHAIN-CODE>",
    ),
    raw_weights: List[str] = typer.Option(
        [],
        "-w",
        "--weight",
        help="weights for rdcs as a comma separated triple of 2 atom names and a weight [no spaces e.g HN,N,1.0],"
        " multiple weights may be added by repeated calls, efault is HN,N,1.0",
    ),
    frame_selectors: List[str] = typer.Option(
        [],
        "-f",
        "--frame",
        help="selector for the rdc restraint frame to use, can be called multiple times and include wild cards",
    ),
):
    """- convert nef rdc restraints to pales"""

    if not chains:
        chains = ["A"]
    weights = _parse_weights(raw_weights)
    sequence = get_sequence_or_exit()

    sequence_3_let = sequence_residues_to_sequence_3let(sequence)

    sequence_1_let = translate_3_to_1(sequence_3_let)

    NEF_RDC_CATEGORY = "nef_rdc_restraint_list"

    entry = create_entry_from_stdin_or_exit()
    rdc_frames = entry.get_saveframes_by_category(NEF_RDC_CATEGORY)

    if frame_selectors is []:
        frame_selectors = [""]
    frames = select_frames_by_name(rdc_frames, frame_selectors)

    restaints = _rdc_restraints_from_frames(frames, chains, weights)

    restaints = _translate_atom_names(restaints, "xplor")

    print(f'REMARK NEF CHAIN {", ".join(chains)}')
    print(f"REMARK NEF START RESIDUE {sequence[0].sequence_code}")
    print()
    print_pipe_sequence(sequence_1_let)
    print()
    _print_restraints(restaints, weights)


def _translate_atom_names(
    restraints: list[RdcRestraint], naming_scheme="iupac"
) -> List[RdcRestraint]:
    result = []
    for restraint in restraints:
        atom_1 = restraint.atom_1
        atom_2 = restraint.atom_2

        if atom_1.atom_name == "H":
            restraint.atom_1 = replace(atom_1, atom_name="HN")

        if atom_2.atom_name == "H":
            restraint.atom_1 = replace(atom_2, atom_name="HN")

        result.append(restraint)

    return result


def _rdc_restraints_from_frames(
    frames: List[Saveframe], chains: List[str], weights: Dict[Tuple[str, str], float]
):
    result = []
    for frame in frames:
        for row in loop_row_dict_iter(frame.loops[0]):

            atom_1 = AtomLabel(
                SequenceResidue(
                    row["chain_code_1"],
                    int(row["sequence_code_1"]),
                    row["residue_name_1"],
                ),
                row["atom_name_1"],
            )
            atom_2 = AtomLabel(
                SequenceResidue(
                    row["chain_code_2"],
                    int(row["sequence_code_2"]),
                    row["residue_name_2"],
                ),
                row["atom_name_2"],
            )

            weight_key = _build_weights_key(row["atom_name_1"], row["atom_name_2"])
            # this should be
            weight = weights[weight_key] if weight_key in weights else 1.0
            rdc = RdcRestraint(
                atom_1,
                atom_2,
                row["target_value"],
                row["target_value_uncertainty"],
                weight,
            )

            result.append(rdc)

    return sorted(result)


def _parse_weights(raw_weights):

    result = {}
    for raw_weight in raw_weights:

        raw_weight_fields = raw_weight.split(",")

        if len(raw_weight_fields) != 3:
            exit_error(
                f"bad weight {raw_weight} weights should have 3 fields separated by commas [no spaces]"
                f" atom_1,atom_2,weight. e.g HN,N,1.0"
            )

        atom_1, atom_2, raw_rdc_weight = raw_weight_fields

        rdc_weight = read_float_or_exit(
            raw_rdc_weight,
            message=f"weights should have 3 fields separated by commas [no spaces] atom_1,atom_2,weight weight should"
            f" be a float i got {raw_weight} for field feild [{raw_rdc_weight}] which is not a float",
        )

        key = _build_weights_key(atom_1, atom_2)
        result[key] = rdc_weight

    # it not possibe to add default weights?
    HN_N_key = _build_weights_key("HN", "N")
    if HN_N_key not in result:
        result[HN_N_key] = 1.0

    return result


def _build_weights_key(atom_1, atom_2):
    return tuple(sorted([atom_1, atom_2]))


def _print_restraints(
    restraints: List[RdcRestraint], weights: Dict[Tuple[str, str], float]
):
    VARS = "VARS   RESID_I RESNAME_I ATOMNAME_I RESID_J RESNAME_J ATOMNAME_J D DD W".split()
    FORMAT = "FORMAT %5d     %6s       %6s        %5d     %6s    %6s %9.3f %9.3f %.2f".split()

    table = [VARS, FORMAT]

    for i, restraint in enumerate(restraints):
        atom_1 = restraint.atom_1
        atom_2 = restraint.atom_2

        weights_key = _build_weights_key(atom_1.atom_name, atom_2.atom_name)
        weight = weights[weights_key]
        row = [
            "",
            atom_1.residue.sequence_code,
            atom_1.residue.residue_name,
            atom_1.atom_name,
            atom_2.residue.sequence_code,
            atom_2.residue.residue_name,
            atom_2.atom_name,
            restraint.rdc,
            restraint.rdc_error,
            weight,
        ]
        table.append(row)

    print(tabulate.tabulate(table, tablefmt="plain"))



from textwrap import dedent

from freezegun import freeze_time

import pytest
from icecream import ic

from typer.testing import CliRunner

from lib.sequence_lib import translate_1_to_3, BadResidue, sequence_3let_to_sequence_residues
from lib.structures import SequenceResidue
from lib.test_lib import assert_lines_match, isolate_frame, path_in_test_data, run_and_report

MOLECULAR_SYSTEM = 'nef_molecular_system'
METADATA_NMRVIEW ='nef_nmr_meta_data'
NMRVIEW_IMPORT_SEQUENCE = ['nmrview', 'import', 'sequence']

runner = CliRunner()


@pytest.fixture
def using_nmrview():
    # register the module under test
    import transcoders.nmrview


EXPECTED_3AA = '''\
save_nef_molecular_system
   _nef_molecular_system.sf_category   nef_molecular_system
   _nef_molecular_system.sf_framecode  nef_molecular_system

   loop_
      _nef_sequence.index
      _nef_sequence.chain_code
      _nef_sequence.sequence_code
      _nef_sequence.residue_name
      _nef_sequence.linking
      _nef_sequence.residue_variant
      _nef_sequence.cis_peptide

     1   A   1   ALA   start    .   .
     2   A   2   ALA   middle   .   .
     3   A   3   ALA   end      .   .

   stop_

save_'''

EXPECTED_3AA10 = '''\
save_nef_molecular_system
   _nef_molecular_system.sf_category   nef_molecular_system
   _nef_molecular_system.sf_framecode  nef_molecular_system

   loop_
      _nef_sequence.index
      _nef_sequence.chain_code
      _nef_sequence.sequence_code
      _nef_sequence.residue_name
      _nef_sequence.linking
      _nef_sequence.residue_variant
      _nef_sequence.cis_peptide

     1   A   10   ALA   start    .   .
     2   A   11   ALA   middle   .   .
     3   A   12   ALA   end      .   .

   stop_

save_'''


# noinspection PyUnusedLocal
def test_3aa(typer_app, using_nmrview, monkeypatch):

    monkeypatch.setattr('sys.stdin.isatty', lambda: False)

    path = path_in_test_data(__file__, '3aa.seq')
    result = run_and_report(typer_app, [*NMRVIEW_IMPORT_SEQUENCE, path], input=HEADER)

    mol_sys_result = isolate_frame(result.stdout, '%s' % MOLECULAR_SYSTEM)

    assert_lines_match(EXPECTED_3AA, mol_sys_result)


# noinspection PyUnusedLocal
def test_3aa10(typer_app, using_nmrview, monkeypatch):

    monkeypatch.setattr('sys.stdin.isatty', lambda: False)

    path = path_in_test_data(__file__, '3aa10.seq')
    result = run_and_report(typer_app, [*NMRVIEW_IMPORT_SEQUENCE, path], input=HEADER)

    mol_sys_result = isolate_frame(result.stdout, '%s' % MOLECULAR_SYSTEM)

    assert_lines_match(EXPECTED_3AA10, mol_sys_result)


HEADER = open(path_in_test_data(__file__,'test_header_entry.txt', local=False)).read()

EXPECTED_HEADER = '''\
save_nef_nmr_meta_data
   _nef_nmr_meta_data.sf_category      nef_nmr_meta_data
   _nef_nmr_meta_data.sf_framecode     nef_nmr_meta_data
   _nef_nmr_meta_data.format_name      nmr_exchange_format
   _nef_nmr_meta_data.format_version   1.1
   _nef_nmr_meta_data.program_name     NEFPipelines
   _nef_nmr_meta_data.script_name      util.py
   _nef_nmr_meta_data.program_version  0.0.1
   _nef_nmr_meta_data.creation_date    2012-01-14T12:00:01.123456
   _nef_nmr_meta_data.uuid             NEFPipelines-2012-01-14T12:00:01.123456-1043321819
   _nef_nmr_meta_data.creation_time    2012-01-14T12:00:01.123456

   loop_
      _nef_run_history.run_number
      _nef_run_history.program_name
      _nef_run_history.program_version
      _nef_run_history.script_name

     1   NEFPipelines   0.0.1   header.py

   stop_

save_
'''


# noinspection PyUnusedLocal
@freeze_time("2012-01-14 12:00:01.123456")
def test_pipe_header(typer_app, using_nmrview, monkeypatch, fixed_seed):

    monkeypatch.setattr('sys.stdin.isatty', lambda: True)

    path = path_in_test_data(__file__, '3aa10.seq')
    path_header = path_in_test_data(__file__,'test_header_entry.txt',local=False)
    result = run_and_report(typer_app, [*NMRVIEW_IMPORT_SEQUENCE, '--pipe', path_header, path])


    mol_sys_result = isolate_frame(result.stdout, '%s' % MOLECULAR_SYSTEM)
    meta_data_result = isolate_frame(result.stdout, '%s' % METADATA_NMRVIEW)

    assert_lines_match(EXPECTED_3AA10, mol_sys_result)
    assert_lines_match(EXPECTED_HEADER, meta_data_result)


# noinspection PyUnusedLocal
@freeze_time("2012-01-14 12:00:01.123456")
def test_header(typer_app, using_nmrview, monkeypatch, fixed_seed):

    monkeypatch.setattr('sys.stdin.isatty', lambda: False)

    path = path_in_test_data(__file__, '3aa10.seq')
    result = run_and_report(typer_app, [*NMRVIEW_IMPORT_SEQUENCE, path], input=HEADER)

    mol_sys_result = isolate_frame(result.stdout, '%s' % MOLECULAR_SYSTEM)

    assert_lines_match(EXPECTED_3AA10, mol_sys_result)


if __name__ == '__main__':
    pytest.main([f'{__file__}', '-vv'])
